library ArcNeuralNetwork.Layer;

//Imports

import "neuron.dart";
import "network.dart";

/*
Class Layer: represent a layer of the neural network
      Parameters:
        - Id: identifiers the layer --> String
        - Neurons: neurons of the layer --> List<Neuron>
        - Network: network to which the layer belongs --> network
      Methods:
        - Add Neuron
        - Remove Neuron
        - Get Neuron
        - Set Neuron
        - Get Weight
        - Set Weight
        - Calculate Output: get output of all neurons in layer
*/

class Layer {

  String id;
  List<Neuron> neurons;
  Network network;

  //////////////////////////////////////////////
  ////////////// Constructors //////////////////
  //////////////////////////////////////////////

  //Create a simple layer

  Layer(this.id) {
    neurons = [];
  }

  //Create a layer with a neuron array

  Layer.Neurons(this.id, this.neurons) {
  }

  ///////////////////////////////////  Add Neuron  /////////////////////////////////

  void addNeuron(Neuron neuron) {
    if (neuron == null) {
      throw ("Neuron is empty");
    }
    neuron.layer = this;
    this.neurons.add(neuron);
  }

  void addNeuronAt(int index, neuron) {
    if (neuron == null) {
      throw ("Neuron is empty");
    }
    if (index > this.neurons.length) {
      throw ("Index out of bounds");
    }
    neuron.layer = this;
    this.neurons.insert(index, neuron);
  }

  ///////////////////////////////////  Remove Neuron  /////////////////////////////////

  void removeAllNeurons() {
    this.neurons = [];
  }

  void removeNeuronAt(int index) {
    this.neurons.removeAt(index);
  }

  void removeNeuron(Neuron neuron) {
    this.neurons.remove(neuron);
  }

  ///////////////////////////////////  Get Neuron  /////////////////////////////////

  int getNeuronIndexOf(Neuron neuron) {
    return this.neurons.indexOf(neuron);
  }

  Neuron getNeuronByIndex(int index) {
    if (index > this.neurons.length) {
      throw ("Index out of bounds");
    }
    return this.neurons[index];
  }

  Neuron getNeuronById(String id) {
    for (var neuron in this.neurons) {
      if (neuron.id == id) {
        return neuron;
      }
    }
    return null;
  }

  ///////////////////////////////////  Set Neuron  /////////////////////////////////

  void setNeuronAt(int index, Neuron neuron) {
    if (neuron == null) {
      throw ("Neuron is empty");
    }
    if (index > this.neurons.length) {
      throw ("Index out of bounds");
    }
    neuron.layer = this;
    this.neurons[index] = neuron;
  }

  ///////////////////////////////////  Check Neuron  /////////////////////////////////

  bool get hasNeurons {
    return this.neurons.isEmpty;
  }

  ///////////////////////////////////  Get Weight  /////////////////////////////////

  List<List<Weight>> get weights {
    List<List<Weight>> weights = [];
    for (Neuron neuron in this.neurons) {
      weights.add(neuron.getAllWeights());
    }
    return weights;
  }

  List<List<double>> get weightsValues {
    List<List<double>> weights = [];
    for (Neuron neuron in this.neurons) {
      weights.add(neuron.getAllWeightsValues());
    }
    return weights;
  }

  List<List<double>> get weightsPreviousValues {
    List<List<double>> weights = [];
    for (Neuron neuron in this.neurons) {
      weights.add(neuron.getAllWeightsPreviousValues());
    }
    return weights;
  }

  ///////////////////////////////////  Set weight  /////////////////////////////////

  void setAllWeightValueRandom([double minRange, double maxRange]) {
    for (Neuron neuron in this.neurons) {
      neuron.setAllWeightValueRandom(minRange, maxRange);
    }
  }

  void setAllWeightValue(double value) {
    for (Neuron neuron in this.neurons) {
      neuron.setAllWeightValue(value);
    }
  }

  ///////////////////////////////////  Calculate output of all neurons of the layer  /////////////////////////////////

  void calculateOutput() {
    for (Neuron neuron in this.neurons) {
      neuron.neuronOutput();
    }
  }
}
