library ArcNeuralNetwork.Neuron;

//Imports

import "dart:math";
import "layer.dart";
import 'Functions/inputfunction.dart';
import 'Functions/activationfunction.dart';

/*
Class Neuron: represent a neuron
      Parameters:
        - Id: identifier for the neuron --> String
        - InputConnection: a list of all the input connections --> List<Connection>
        - OutputConnections: a list of all output connections --> List<Connection>
        - Input: value of the input --> double
        - Output: value of the output --> double
        - Error: local error of the neuron --> double
        - Activation function --> ActivationFunction
        - Input Function: to calculate the input value --> InputFunction
        - Layer: layer to which the neuron belongs --> Layer
      Methods:
        - neuronOutput: calculates the output of the neuron in base to the input and activation function.
        - Check connections: boolean methods to know if there are connections set.
        - Add connection: add connections to neuron.
        - Remove connection: remove connection from neuron.
*/

class Neuron {

  String id;
  List<Connection> inputConnection;
  List<Connection> outputConnection;
  double input = 0.0;
  double output = 0.0;
  double error = 0.0;
  ActivationFunction activationFunction;
  InputFunction inputFunction;
  Layer layer;

  //////////////////////////////////////////////
  ////////////// Constructors //////////////////
  //////////////////////////////////////////////

  //Create a simple neuron with a stp activation function and a weight combination function for the input

  Neuron(this.id) {
    this.inputFunction = new WeightCombination();
    this.activationFunction = new Step();
    this.inputConnection = [];
    this.outputConnection = [];
  }

  //Create a simple neuron with custom activation function and input function

  Neuron.Functions(this.id, this.inputFunction, this.activationFunction) {
    this.inputConnection = [];
    this.outputConnection = [];
  }

  //Methods

  //calculates the output of the neuron in base to the input and activation function

  void neuronOutput() {
    if ((this.inputConnection.length > 0)) {
      this.input = this.inputFunction.getOutput(this.inputConnection);
    }
    this.output = this.activationFunction.getOutput(this.input);
  }

  ///////////////////////////////////  Check connections  /////////////////////////////////

  bool get hasInputConnection {
    if (this.inputConnection.length > 0) {
      return true;
    } else {
      return false;
    }
  }

  bool get hasOutputConnection {
    if (this.outputConnection.length > 0) {
      return true;
    } else {
      return false;
    }
  }

  bool hasInputConnectionFrom(Neuron neuron) {
    return this.inputConnection.contains(neuron);
  }

  bool hasOutputConnectionTo(Neuron neuron) {
    return this.outputConnection.contains(neuron);
  }

  ///////////////////////////////////  Adding connections  /////////////////////////////////

  addInputConnection(Connection connection) {
    if (connection == null) {
      throw ("Attempt to add null connection to neuron!");
    }

    // make sure that connection instance is pointing to this neuron
    if (connection.to != this) {
      throw ("Cannot add input connection - bad fromNeuron specified!");
    }

    // if this neuron is already connected to neuron specified in connection do nothing
    if (this.hasInputConnectionFrom(connection.from)) {
      return;
    }
    this.inputConnection.add(connection);
    Neuron fromNeuron = connection.from;
    fromNeuron.addOutputConnection(connection);
  }

  addOutputConnection(Connection connection) {

    if (connection == null) {
      throw ("Attempt to add null connection to neuron!");
    }

    // make sure that connection instance is pointing to this neuron
    if (connection.from != this) {
      throw ("Cannot add output connection - bad fromNeuron specified!");
    }

    // if this neuron is already connected to neuron specified in connection do nothing
    if (this.hasOutputConnectionTo(connection.to)) {
      return;
    }
    this.outputConnection.add(connection);
  }

  ///////////////////////////////////  Remove connections  /////////////////////////////////

  removeInputConnection(Connection connection) {
    this.inputConnection.remove(connection);
  }

  removeOutputConnection(Connection connection) {
    this.outputConnection.remove(connection);
  }

  removeInputConnectionWithNeuronFrom(Neuron neuron) {
    Connection connection = this.getInputConnectionWithNeuronFrom(neuron);
    this.removeInputConnection(connection);
  }

  removeOutputConnectionWithNeuronTo(Neuron neuron) {
    Connection connection = this.getOutputConnectionWithNeuronTo(neuron);
    this.removeOutputConnection(connection);
  }

  removeAllInputConnections() {
    this.inputConnection = [];
  }

  removeAllOutputConnections() {
    this.outputConnection = [];
  }

  removeAllConnections() {
    removeAllInputConnections();
    removeAllOutputConnections();
  }

  ///////////////////////////////////  Get input connections  ////////////////////////////////////

  Connection getInputConnection(Connection connection) {
    for (var con in this.inputConnection) {
      if (con == connection) return connection;
    }
    return null;
  }

  int getInputConnectionIndexOf(Connection connection) {
    return this.inputConnection.indexOf(connection);
  }

  Connection getInputConnectionWithNeuronFrom(Neuron neuron) {
    for (var connection in this.inputConnection) {
      if (connection.from == neuron) return connection;
    }
    return null;
  }

  Connection getInputConnectionWithWeight(Weight Weight) {
    for (var connection in this.inputConnection) {
      if (connection.weight == Weight) return connection;
    }
    return null;
  }

  ///////////////////////////////////  Get output connections  ////////////////////////////////////

  Connection getOutputConnection(Connection connection) {
    for (var con in this.outputConnection) {
      if (con == connection) return connection;
    }
    return null;
  }

  int getOutputConnectionIndexOf(Connection connection) {
    return this.outputConnection.indexOf(connection);
  }

  Connection getOutputConnectionWithWeight(Weight Weight) {
    for (var connection in this.outputConnection) {
      if (connection.weight == Weight) return connection;
    }
    return null;
  }

  Connection getOutputConnectionWithNeuronTo(Neuron neuron) {
    for (var connection in this.outputConnection) {
      if (connection.to == neuron) return connection;
    }
    return null;
  }

  ///////////////////////////////////  Get weights /////////////////////////////////////////

  Weight getWeight(Connection connection) {
    return this.getInputConnection(connection).weight;
  }

  double getWeightValue(Connection connection) {
    return this.getInputConnection(connection).weightValue;
  }

  double getWeightPreviousValue(Connection connection) {
    return this.getInputConnection(connection).weightPreviousValue;
  }

  List<Weight> getAllWeights() {
    List<Weight> Weights = [];
    for (var connection in this.inputConnection) {
      Weights.add(connection.weight);
    }
    return Weights;
  }

  List<double> getAllWeightsValues() {
    List<double> Weights = [];
    for (var connection in this.inputConnection) {
      Weights.add(connection.weight.value);
    }
    return Weights;
  }

  List<double> getAllWeightsPreviousValues() {
    List<double> Weights = [];
    for (var connection in this.inputConnection) {
      Weights.add(connection.weight.previousValue);
    }
    return Weights;
  }

  ///////////////////////////////////  Set weights  ////////////////////////////////////

  void setWeight(Connection connection, Weight Weight) {
    this.getInputConnection(connection).weight = Weight;
  }

  void setWeightValue(Connection connection, double value) {
    this.getInputConnection(connection).weightValue = value;
  }

  void setWeightValueRandom(Connection connection, [double rangeMin, double rangeMax]) {
    this.getInputConnection(connection).weight.weightRandom(rangeMin, rangeMax);
  }

  void setAllWeight(List<Weight> Weights) {
    if (Weights.length == this.inputConnection.length) {
      for (var i = 0; i < this.inputConnection.length; i++) {
        this.inputConnection[i].weight = Weights[i];
      }
    } else {
      throw ("Weights length must be the same as connections!!!");
    }
  }

  void setAllWeightValue(double value) {
    for (var i = 0; i < this.inputConnection.length; i++) {
      this.inputConnection[i].weight.value = value;
    }
  }

  void setAllWeightValueRandom([double rangeMin, double rangeMax]) {
    for (var i = 0; i < this.inputConnection.length; i++) {
      this.inputConnection[i].weight.weightRandom(rangeMin, rangeMax);
    }
  }
}

/*
Class Connection: represent a connection between two neurons
      Parameters:
        - From: neuron where connection starts --> Neuron
        - To: neuron where connection ends -->Neuron
        - Weight: weight associated to the connection -->Weight
      Methods:
        - inputValue: calculates the input of the connection. This is important because that input value is used to calculate the total input of a neuron.
*/

class Connection {

  Neuron from;
  Neuron to;
  Weight weight;

  //////////////////////////////////////////////
  ////////////// Constructors //////////////////
  //////////////////////////////////////////////

  //Create a connection between two neurons with a random weight

  Connection(this.from, this.to) {
    this.weight = new Weight();
  }

  //Create a connection between two neurons with a specific weight

  Connection.Weight(this.from, this.to, this.weight);

  //Create a connection between two neurons with a weight with a specific value

  Connection.WeightValue(this.from, this.to, double value) {
    this.weight = new Weight.Value(value);
  }

  //Create a connection between two neurons with a random range value

  Connection.WeightRandomRange(this.from, this.to, double rangeMin, double rangeMax) {
    this.weight = new Weight();
    this.weight.weightRandom(rangeMin, rangeMax);
  }

  //Getters and Setters

  //Get de input value of a connection

  double get inputValue => this.from.output;

  double get weightValue => this.weightValue;

  set weightValue(double value) => this.weightValue = value;

  double get weightPreviousValue => this.weightPreviousValue;

  set weightPreviousValue(double value) => this.weightPreviousValue = value;

  double get weightVariation => this.weightVariation;

}

/*
Class Weight: represent the weight of a connection
      Parameters:
        - Value: the value of the weight at this moment -->double
        - PreviousValue: the value of the weight at the moment before -->double
        - Variation: the difference between the value and the previous value
      Methods:
        - weightRandom: generates a random number or a random number inside a range
        - weightIncrement: increments the value of the weight and calculates the variation.
        - weightDecrement: decrements the value of the weight and calculates the variation.
*/

class Weight {

  double value;
  double previousValue;
  double variation;

  //////////////////////////////////////////////
  ////////////// Constructors //////////////////
  //////////////////////////////////////////////

  //Create a random weight

  Weight() {
    this.weightValue = weightRandom();
    this.variation = 0.0;
  }

  //Create a weight with specific value

  Weight.Value(this.value) {
    this.variation = 0.0;
  }

  //Getters and Setters

  set weightValue(double value) {
    if (this.value = null) {
      this.value = value;
    } else {
      this.previousValue = this.value;
      this.value = value;
      this.variation = (this.value - this.previousValue).abs();
    }
  }

  set weightPreviousValue(double value) => this.previousValue = value;

  set weightVariation(double value) => this.variation = value;

  double get weightValue => this.value;

  double get weightPreviousValue => this.previousValue;

  double get weightVariation => this.variation;

  //Get random value

  double weightRandom([double rangeMin, double rangeMax]) {
    Random r = new Random();
    if (rangeMax != null && rangeMin != null) {
      return rangeMin + (rangeMax - rangeMin) * r.nextDouble();
    } else {
      return r.nextDouble();
    }
  }

  //Increment and decrement weight value

  void weightIncrement(double value) {
    this.previousValue = this.value;
    this.value = this.value + value;
    this.variation = (this.previousValue - this.value).abs();

  }

  void weightDecrement(double value) {
    this.previousValue = this.value;
    this.value = this.value - value;
    this.variation = (this.previousValue - this.value).abs();
  }

}
